//@ts-check

import JadefinIntegrity from '../../JadefinIntegrity.js';

import Jadefin from "../../Jadefin.js";
import JadefinMod from "../../JadefinMod.js";
import JadefinModules from "../../JadefinModules.js";
import JadefinUtils from "../../JadefinUtils.js";

import { rd, rdom, rd$, RDOMListHelper } from "../../utils/rdom.js";

export class FilterworksFilter {
    name = "Hello, World!";

    /**
     * @param {HTMLVideoElement} video
     */
    start(video) {
    }

    stop(video) {
    }
}

export class FilterworksSrc {
    modName = "ExampleSource";

    /** @type {{
        [id: string]: FilterworksFilter;
    }} */
    filters = {
        test: new FilterworksFilter()
    };
}

export default JadefinIntegrity("Filterworks", import.meta.url, () => new (class Filterworks extends JadefinMod {
    PATH_OFF = "Filterworks/Off";

    _sources = [];
    _pathsToFilters = {};
    _filtersToMeta = {};

    /** @type {any} */
    appliedFilter = null;

    canvasParentEl = rd$()`
        <div class="filterworks-canvasParent">
        </div>
    `;

    /** @type {typeof this._sources} */
    sources = new Proxy(this._sources, {
        deleteProperty: (target, property) => {
            delete target[property];

            this.updateSources();

            return true;
        },

        set: (target, property, value, receiver) => {
            target[property] = value;

            this.updateSources();

            return true;
        }
    });

    constructor() {
        super();

        this.sources.push({
            modName: "Filterworks",
            filters: {
                Off: {
                    name: "Off",
                    start: (video) => {},
                    stop: () => {}
                }
            }
        });
    }

    get currentFilterPathPath() {
        let id = "default";

        const item = JadefinUtils.currentPlayer?.streamInfo?.item;
        if (item) {
            id = item.SeriesId || item.ParentId || item.Id || id;
        }

        return `currentFilterPath-${id}`;
    }

    get currentFilterPath() {
        return this.storage.get(this.currentFilterPathPath, this.PATH_OFF);
    }

    set currentFilterPath(value) {
        this.storage.set(this.currentFilterPathPath, value);
        this.update();
    }

    get currentFilter() {
        return this._pathsToFilters[this.currentFilterPath] || this._pathsToFilters[this.PATH_OFF];
    }

    async init(name, url) {
        await super.init(name, url);

        const ExtrasMenu = /** @type {import("../ExtrasMenu.js").default} */ (Jadefin.getMod("ExtrasMenu"));

        await this.initStyle();

        ExtrasMenu.items.push({
            name: "Filter",
            in: ExtrasMenu.IN_CUSTOM,
            inCustom: (current, item) => {
                item.asideText = this.currentFilter.name;

                return (current & (ExtrasMenu.IN_PLAYBACKSETTINGS)) == current;
            },
            cb: async (positionTo) => {
                const items = [];

                items.push({id: this.PATH_OFF, name: this._pathsToFilters[this.PATH_OFF].name, selected: this.currentFilter == this._pathsToFilters[this.PATH_OFF]});

                for (const path in this._pathsToFilters) {
                    if (path == this.PATH_OFF) {
                        continue;
                    }

                    items.push({id: path, name: this._pathsToFilters[path].name, selected: this.currentFilterPath == path});
                }

                JadefinModules.actionSheet.show({
                    positionTo,
                    items
                }).then(id => {
                    if (!id) {
                        return;
                    }

                    this.currentFilterPath = id;
                });
            }
        });

        document.addEventListener("viewshow", () => {
            if (JadefinUtils.routePathIsVideo) {
                if (this._tryUpdateRepeat) {
                    clearInterval(this._tryUpdateRepeat);
                }

                clearInterval(this._tryUpdateRepeat);
                this._tryUpdateRepeat = setInterval(() => this.tryUpdate(), 100);
            }

            this.update();
        });

        this.log.i("Ready");
    }

    updateSources() {
        this._pathsToFilters = {};

        for (const src of this._sources) {
            for (const id in src.filters) {
                this._pathsToFilters[`${src.modName}/${id}`] = src.filters[id];
                this._filtersToMeta[src.filters[id]] = { src, id };
            }
        }
    }

    tryUpdate() {
        const video = JadefinUtils.video;
        if (!video) {
            return;
        }

        this.update();

        if (this._tryUpdateRepeat) {
            clearInterval(this._tryUpdateRepeat);
        }
    }

    update() {
        const video = JadefinUtils.video;
        if (!video) {
            if (this.appliedFilter) {
                this.appliedFilter.stop();
                this.appliedFilter = null;
            }

            this.video = null;
            this.canvas = null;
            return;
        }

        if (this.video == video && this.appliedFilter == this.currentFilter) {
            return;
        }

        this.appliedFilter?.stop();
        if (this.canvas) {
            video.removeAttribute("data-filterworks-canvas");
            this.canvas.remove();
        }

        this.video = video;

        this.appliedFilter = this.currentFilter;
        this.canvas = this.appliedFilter.start(video);
        if (this.canvas) {
            this.canvas.classList.add("filterworks-canvas");
            this.canvasParentEl.append(this.canvas);

            if (this.canvasParentEl.parentElement != video.parentElement) {
                video.parentElement?.prepend(this.canvasParentEl);
            }

            video.setAttribute("data-filterworks-canvas", "1");
        }
    }
})());
